/// ReWindowLibrary
/// A C++20 Library for creating and managing windows in a platform-independent manner
/// Developed and Maintained by the boys @ Redacted Software.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file keyboard.h
/// @desc A class that models the functionality of a mouse / pointer device.
/// @edit 2024-07-29

#pragma once

//#include <rewindow/types/inputdevice.hpp>
#include <rewindow/types/mousebutton.h>
#include <rewindow/types/key.h>
#include <Event.h>
#include <map>
#include "rewindow/data/scancode.h"
#include <rewindow/types/inputdevice.hpp>
//#include <rewindow/types/keyboard.h>
//#include <rewindow/types/gamepad.h>

namespace ReWindow
{

    //using Key = ReScanCode;

    using Button = ReScanCode;

    /*
    enum class DeviceType {
        MOUSE, KEYBOARD, GAMEPAD, TOUCHSCREEN, REMOTE, VRHEADSET
    };

    enum class DeviceState {
        OK, LOW_BATTERY, UNAVAILABLE
    };
     */


    /*
    class InputDeviceImpl { };
    class InputDeviceImplLinux { };
     */
    class LinuxMouseImpl { };
    class LinuxKeyboardImpl {};

    // TODO: Remember to break InputDevice into it's own file and not define it twice!!!
    /*
    class InputDevice {
    public:
        virtual void Update(double elapsed){}

        /// Returns the device's index among devices of the same type.
        uint GetIndex() const;

        virtual DeviceType GetType() const = 0;
        DeviceState GetDeviceState() const;
        bool IsAnyButtonDown();
    };
     */

    /*
    struct BinaryActuatorState
    {
        double TimeHeld;
        bool Pressed;
    };
     */

    using ButtonState = BinaryActuatorState;
    //using KeyState = BinaryActuatorState;

    /*
    struct KeyboardState
    {
        std::map<Key, KeyState> Keys;

    };
     */

    /*class MouseState {
    public:
        struct
        {
            bool LMB = false;
            bool RMB = false;
            bool MMB = false;
            bool SideButton1 = false;
            bool SideButton2 = false;
            bool MWheelUp = false;
            bool MWheelDown = false;
        } Buttons;


        Vector2 Position;
        int Wheel = 0;

        [[nodiscard]] bool IsDown(const MouseButton& btn) const
        {
            if (btn == MouseButtons::Left) return Buttons.LMB;
            if (btn == MouseButtons::Right) return Buttons.RMB;
            if (btn == MouseButtons::Middle) return Buttons.MMB;
            if (btn == MouseButtons::Mouse4) return Buttons.SideButton1;
            if (btn == MouseButtons::Mouse5) return Buttons.SideButton2;
            //if (btn == MouseButtons::MWheelUp) return Buttons.MWheelUp;
            //if (btn == MouseButtons::MWheelDown) return Buttons.MWheelDown;

            return false; // Unknown button?
        }

        void Set(const MouseButton& btn, bool state)
        {
            if (btn == MouseButtons::Left) Buttons.LMB = state;
            if (btn == MouseButtons::Right) Buttons.RMB = state;
            if (btn == MouseButtons::Middle) Buttons.MMB = state;
            if (btn == MouseButtons::Mouse4) Buttons.SideButton1 = state;
            if (btn == MouseButtons::Mouse5) Buttons.SideButton2 = state;
            //if (btn == MouseButtons::MWheelUp) Buttons.MWheelUp = state;
            //if (btn == MouseButtons::MWheelDown) Buttons.MWheelDown = state;
        }

        bool& operator[](const MouseButton& btn)
        {
            if (btn == MouseButtons::Left) return Buttons.LMB;
            if (btn == MouseButtons::Right) return Buttons.RMB;
            if (btn == MouseButtons::Middle) return Buttons.MMB;
            if (btn == MouseButtons::Mouse4) return Buttons.SideButton1;
            if (btn == MouseButtons::Mouse5) return Buttons.SideButton2;
            //if (btn == MouseButtons::MWheelUp) return Buttons.MWheelUp;
            //if (btn == MouseButtons::MWheelDown) return Buttons.MWheelDown;

            throw std::runtime_error("Attempted to handle unmapped mouse button");
        }
    };*/

    struct MouseState
    {
        ButtonState Left;
        ButtonState Right;
        ButtonState Middle;
        ButtonState Mouse4;
        ButtonState Mouse5;
        ButtonState WheelUp;
        ButtonState WheelDown;

        Vector2 Position;
        Vector2 Delta;
        Vector2 Velocity;
        int Wheel;


    };


    struct ThumbstickState
    {

        Vector2 Unit;
    };

    class Thumbstick
    {

    };

    struct TriggerState
    {

    };

    class Trigger
    {

    };

    struct GamepadState
    {
        ButtonState X;
        ButtonState Y;
        ButtonState A;
        ButtonState B;

        // TODO: Also have analog DPad;
        ButtonState DPadLeft;
        ButtonState DPadRight;
        ButtonState DPadUp;
        ButtonState DPadDown;

        ButtonState Pause;
        ButtonState Select;
        ButtonState Home;

        ButtonState LeftThumbButton;
        ButtonState RightThumbButton;

        TriggerState LeftTrigger;
        TriggerState RightTrigger;

        ThumbstickState LeftThumbstick;
        ThumbstickState RightThumbstick;

    };

    class TouchscreenState
    {

    };

    /*
     *  Gamepad
     */

    constexpr int MaxDeviceStateHistory = 100;

    class Gamepad : public InputDevice
    {
    public:
        uint GetButtonCount() const;
        uint GetRumbleMotorCount() const;
    protected:
        Thumbstick LeftThumb;
        Thumbstick RightThumb;

        Trigger RightTrigger;
        Trigger LeftTrigger;

        GamepadState current_state;
        GamepadState previous_state;
        GamepadState history[MaxDeviceStateHistory];
    };


    class XboxGamepad : public Gamepad {};
    class PS4Gamepad : public Gamepad {};

    /*
     *  Pointer
     */

    class PointerDevice : public InputDevice {
    public:
        void Update(double elapsed) override {
            InputDevice::Update(elapsed);
        }
    };

    /*
    struct InputEvent
    {
        std::chrono::system_clock::time_point Timestamp;
    };
     */

    struct MouseButtonEvent : public InputEvent
    {

    };

    class MouseMoveEvent
    {

    };

    class MouseWheelEvent
    {

    };

    struct TouchEvent
    {

        int TouchID;
        Vector2 Position;
        Vector2 Velocity;
        float Pressure;
    };

    class Mouse : public PointerDevice
    {
    public:
    public:
        Event<> Moved;
        Event<> ButtonPressed;
        Event<> ButtonReleased;
        Event<> WheelMoved;
    public:
        Mouse();
        ~Mouse();
        Vector2 GetPosition() const;
        bool IsButtonDown(const MouseButton& btn) const;
        ButtonState GetButtonState(const MouseButton& btn) const;
        void Update(double elapsed) override;
        DeviceType GetType() const override;
        void DoButtonPress(const MouseButton& btn);
        void DoButtonRelease(const MouseButton& btn);
        void DoMove(const Vector2& pos);
        void DoScrollWheel(int scroll);
    protected:
        void SetButtonPressed(const MouseButton& btn, bool pressed);
        void SetButtonState(const MouseButton& btn, ButtonState state);
    protected:
        MouseState current_state;
        MouseState previous_state;
        MouseState history[MaxDeviceStateHistory];
    public:
    protected:
    private:
    };

    class Touchscreen : public PointerDevice
    {
    public:
        static Event<const TouchEvent&> TouchMoved;
        static Event<const TouchEvent&> TouchPressed;
        static Event<const TouchEvent&> TouchReleased;

        static Touchscreen GetCurrentDevice();

        int GetPosition(int touchIndex);
        int GetTouchCount();
        int GetMaxTouchCount();
        std::vector<int> GetTouches();

        /// Returns the pressure of the touch-press. Most touch screens aren't pressure sensitive, in which case the pressure will be 1.
        float GetPressure(int touchIndex);

    };

    class VRHeadset : public InputDevice { };

    class GenericTVRemote : public InputDevice { };





}