#include <iostream>
#include <rewindow/types/window.h>
#include <rewindow/types/display.h>
#include <rewindow/logger/logger.h>

//aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa Windows :/
#if _WIN32
	#include <windows.h>
#endif

#include <GL/gl.h>

Vector2 mouse_pos;

// TODO: Move to J3ML::LinearAlgebra::Vector2
std::ostream& operator<<(std::ostream& os, const Vector2& v) {
	return os << "{" << v.x << ", " << v.y << "}";
}

class MyWindow : public ReWindow::RWindow {
	public:
		MyWindow(const std::string& title, int w, int h) : ReWindow::RWindow(title, w, h) {}

	void OnMouseMove(const ReWindow::MouseMoveEvent& e) override {}

	void OnKeyDown(const ReWindow::KeyDownEvent& e) override {}

	void OnRefresh(float elapsed) override {
            glClearColor(255, 0, 0, 255);
            glClear(GL_COLOR_BUFFER_BIT);
        GLSwapBuffers();
		auto pos = GetMouseCoordinates();
		//std::cout << pos.x << ", " << pos.y << std::endl;

        //if (IsMouseButtonDown(MouseButtons::MWheelUp))
            //std::cout << "WheelUp Mouse Button" << std::endl;

        //if (IsMouseButtonDown(MouseButtons::MWheelDown))
            //std::cout << "WheelDown Mouse Button" << std::endl;

        if (IsMouseButtonDown(MouseButtons::Left))
            std::cout << "Left Mouse Button" << std::endl;

        if (IsMouseButtonDown(MouseButtons::Right))
            std::cout << "Right Mouse Button" << std::endl;

        if (IsMouseButtonDown(MouseButtons::Middle))
            std::cout << "Middle Mouse Button" << std::endl;


        if (IsMouseButtonDown(MouseButtons::Mouse4))
            std::cout << "Mouse4 Mouse Button" << std::endl;

        if (IsMouseButtonDown(MouseButtons::Mouse5))
            std::cout << "Mouse5 Mouse Button" << std::endl;

        RWindow::OnRefresh(elapsed);
	}

	bool OnResizeRequest(const ReWindow::WindowResizeRequestEvent& e) override {
		std::cout << "resized to " << e.Size.x << ", " << e.Size.y << std::endl;
		return true;
	}

    void OnMouseButtonDown(const ReWindow::MouseButtonDownEvent &e) override
    {
            std::cout << "Overload Down: " << e.Button.Mnemonic << std::endl;
            RWindow::OnMouseButtonDown(e);
    }

    void OnMouseButtonUp(const ReWindow::MouseButtonUpEvent &e) override
    {
            std::cout << "Overload Up: " << e.Button.Mnemonic << std::endl;
            RWindow::OnMouseButtonUp(e);
    }
};

int main() {
	auto* window = new MyWindow("Test Window", 600, 480);
	jlog::Debug(std::format("New window '{}' created. width={} height={}", window->GetTitle(), window->GetWidth(),
                            window->GetHeight()));

    window->SetRenderer(RenderingAPI::OPENGL);
	jlog::Debug(std::format("Rendering API OPENGL set for window '{}'", window->GetTitle()));

	window->Open();
	jlog::Debug(std::format("Opened window '{}'", window->GetTitle()));

	// TODO: Cannot set flags until after window is open
	// Make this work somehow
	jlog::Debug("TODO: Cannot set flags until after window is open");
    window->SetFullscreen(false);
    window->SetVsyncEnabled(true);
    window->SetResizable(true);
    window->SetCursorVisible(false);

	/*
    std::vector<ReWindow::Display> d = ReWindow::Display::getDisplaysFromWindowManager();
    for (ReWindow::Display& display : d) {

        auto w = display.getDefaultGraphicsMode().getWidth();
        auto h = display.getDefaultGraphicsMode().getHeight();
        auto r = display.getDefaultGraphicsMode().getRefreshRate();

        auto aspect = (float)w/(float)h;

        std::cout << "Default: ";
        std::cout << std::format("Mode ({},{}) @ {}hz Ratio {}", w, h, r, aspect) << std::endl;

        for (ReWindow::FullscreenGraphicsMode& fgm : display.getGraphicsModes()) {
            w =  fgm.GetWidth();
            h = fgm.GetHeight();
            r = fgm.getRefreshRate();

            auto aspect = (float)w/(float)h;

            std::cout << std::format("Mode ({},{}) @ {}hz Ratio {}", w, h, r, aspect) << std::endl;
        }
    }
    */

	ReWindow::Logger::Debug(std::format("Window '{}' flags: IN_FOCUS={} FULLSCREEN={} RESIZEABLE={} VSYNC={} QUIT={}",
                                        window->GetTitle(),
                                        window->GetFlag(RWindowFlags::IN_FOCUS),
                                        window->GetFlag(RWindowFlags::FULLSCREEN),
                                        window->GetFlag(RWindowFlags::RESIZABLE),
                                        window->GetFlag(RWindowFlags::VSYNC),
                                        window->GetFlag(RWindowFlags::QUIT)));


	window->OnKeyDownEvent += [&] (ReWindow::KeyDownEvent e) {
        jlog::Debug(e.key.Mnemonic);
	};


	window->OnMouseButtonDownEvent += [&] (ReWindow::MouseButtonDownEvent e) {
		jlog::Debug(e.Button.Mnemonic + std::to_string(e.Button.ButtonIndex));
	};

    window->OnMouseWheelEvent += [&, window] (ReWindow::MouseWheelEvent e)
    {
        std::cout << window->GetMouseWheelPersistent() << std::endl;
    };

	while (window->IsAlive()) {
        window->PollEvents();
        window->ManagedRefresh();
	}
    // Alternatively:
    /*
     * while (window->IsAlive()) {
     *      auto start = window->GetTimestamp();
     *      ourUpdateLogic(window->GetDeltaTime());
     *      window->update();
     *      ourDrawLogic();
     *      window->Refresh();
     *      auto end = window->GetTimestamp();
     *      auto dt = window->ComputeElapsedFrameTimeSeconds(start, end);
     *      window->UpdateFrameTiming(dt);
     * }
     *
     */
}

/*
//Windows :(
#ifdef _WIN32
#ifndef UNICODE
#define UNICODE
#endif

extern "C" {
	int wmain(int argc, wchar_t* argv[]) {
		return main();
	}
}
#endif
*/